@@ -23,7 +23,6 @@
 #include "config.h"
 
 #include <glib.h>
-#include <ctype.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -371,6 +370,35 @@ _pango_ft2_font_map_remove (PangoFontMap *fontmap,
 		       ft2font->font_pattern);
 }
 
+static PangoFT2Family *
+create_family (PangoFT2FontMap *xfontmap,
+	       const char      *family_name)
+{
+  PangoFT2Family *family = g_object_new (PANGO_FT2_TYPE_FAMILY, NULL);
+  family->fontmap = xfontmap;
+  family->family_name = g_strdup (family_name);
+  family->n_faces = -1;
+
+  return family;
+}
+
+static gboolean
+is_alias_family (const char *family_name)
+{
+  switch (family_name[0])
+    {
+    case 'm':
+    case 'M':
+      return (g_ascii_strcasecmp (family_name, "monospace") == 0);
+    case 's':
+    case 'S':
+      return (g_ascii_strcasecmp (family_name, "sans") == 0 ||
+	      g_ascii_strcasecmp (family_name, "serif") == 0);
+    }
+
+  return FALSE;
+}
+
 static void
 pango_ft2_font_map_list_families (PangoFontMap           *fontmap,
 				  PangoFontFamily      ***families,
@@ -379,6 +407,7 @@ pango_ft2_font_map_list_families (PangoFontMap           *fontmap,
   PangoFT2FontMap *ft2fontmap = (PangoFT2FontMap *)fontmap;
   MiniXftFontSet *fontset;
   int i;
+  int count;
 
   if (ft2fontmap->n_families < 0)
     {
@@ -389,9 +418,9 @@ pango_ft2_font_map_list_families (PangoFontMap           *fontmap,
 				  XFT_FAMILY,
 				  NULL);
 
-      ft2fontmap->n_families = fontset->nfont;
-      ft2fontmap->families = g_new (PangoFT2Family *, ft2fontmap->n_families);
+      ft2fontmap->families = g_new (PangoFT2Family *, fontset->nfont + 3); /* 3 standard aliases */
 
+      count = 0;
       for (i = 0; i < fontset->nfont; i++)
 	{
 	  char *s;
@@ -400,12 +429,15 @@ pango_ft2_font_map_list_families (PangoFontMap           *fontmap,
 	  res = MiniXftPatternGetString (fontset->fonts[i], XFT_FAMILY, 0, &s);
 	  g_assert (res == MiniXftResultMatch);
 	  
-	  ft2fontmap->families[i] = g_object_new (PANGO_FT2_TYPE_FAMILY, NULL);
-	  ft2fontmap->families[i]->family_name = g_strdup (s);
-	  ft2fontmap->families[i]->fontmap = ft2fontmap;
-          ft2fontmap->families[i]->n_faces = -1;
+ 	  if (!is_alias_family (s))
+	    ft2fontmap->families[count++] = create_family (ft2fontmap, s);
+
 	}
 
+      ft2fontmap->families[count++] = create_family (ft2fontmap, "Sans");
+      ft2fontmap->families[count++] = create_family (ft2fontmap, "Serif");
+      ft2fontmap->families[count++] = create_family (ft2fontmap, "Monospace");
+
       MiniXftFontSetDestroy (fontset);
     }
   
@@ -500,11 +532,13 @@ pango_ft2_font_map_new_font (PangoFontMap    *fontmap,
   
   if (font)
     {
-      /* Revive fonts from cache */
+      g_object_ref (font);
+      
+      /* Revive font from cache */
       if (font->in_cache)
 	pango_ft2_font_map_cache_remove (fontmap, font);
-      
-      return (PangoFont *)g_object_ref (G_OBJECT(font));
+	  
+      return (PangoFont *)font;
     }
   
   return  (PangoFont *)_pango_ft2_font_new (fontmap, MiniXftPatternDuplicate (match));
@@ -518,17 +552,23 @@ pango_ft2_font_map_load_font (PangoFontMap               *fontmap,
 {
   MiniXftPattern *pattern, *match;
   MiniXftResult res;
-
+  PangoFont *font;
+  
   pattern = pango_ft2_make_pattern (description);
       
   match = MiniXftFontMatch ((Display *)1, 0, pattern, &res);
       
   MiniXftPatternDestroy (pattern);
+
+  font = NULL;
   
   if (match)
-    return  pango_ft2_font_map_new_font (fontmap, match);
+    {
+      font = pango_ft2_font_map_new_font (fontmap, match);
+      MiniXftPatternDestroy (match);
+    }
 
-  return NULL;
+  return font;
 }
 
 static PangoFontset *
@@ -725,6 +765,22 @@ _pango_ft2_font_desc_from_pattern (MiniXftPattern *pattern,
 }
 
 
+static PangoFontDescription *
+make_alias_description (PangoFT2Family *ft2family,
+			gboolean        bold,
+			gboolean        italic)
+{
+  PangoFontDescription *desc = pango_font_description_new ();
+
+  pango_font_description_set_family (desc, ft2family->family_name);
+  pango_font_description_set_style (desc, italic ? PANGO_STYLE_ITALIC : PANGO_STYLE_NORMAL);
+  pango_font_description_set_variant (desc, PANGO_VARIANT_NORMAL);
+  pango_font_description_set_weight (desc, bold ? PANGO_WEIGHT_BOLD : PANGO_WEIGHT_NORMAL);
+  pango_font_description_set_stretch (desc, PANGO_STRETCH_NORMAL);
+
+  return desc;
+}
+
 static PangoFontDescription *
 pango_ft2_face_describe (PangoFontFace *face)
 {
@@ -735,6 +791,18 @@ pango_ft2_face_describe (PangoFontFace *face)
   MiniXftPattern *match_pattern;
   MiniXftPattern *result_pattern;
 
+  if (is_alias_family (ft2family->family_name))
+    {
+      if (strcmp (ft2face->style, "Regular") == 0)
+	return make_alias_description (ft2family, FALSE, FALSE);
+      else if (strcmp (ft2face->style, "Bold") == 0)
+	return make_alias_description (ft2family, TRUE, FALSE);
+      else if (strcmp (ft2face->style, "Italic") == 0)
+	return make_alias_description (ft2family, FALSE, TRUE);
+      else			/* Bold Italic */
+	return make_alias_description (ft2family, TRUE, TRUE);
+    }
+  
   match_pattern = MiniXftPatternBuild (NULL,
 				       XFT_ENCODING, MiniXftTypeString, "iso10646-1",
 				       XFT_FAMILY, MiniXftTypeString, ft2family->family_name,
@@ -827,9 +895,19 @@ _pango_ft2_font_map_get_library (PangoFontMap *fontmap)
 }
 
 /*
- * PangoXFontFamily
+ * PangoFT2FontFamily
  */
+static PangoFT2Face *
+create_face (PangoFT2Family *ft2family,
+ 	     const char     *style)
+{
+  PangoFT2Face *face = g_object_new (PANGO_FT2_TYPE_FACE, NULL);
+  face->style = g_strdup (style);
+  face->family = ft2family;
 
+  return face;
+}
+ 
 static void
 pango_ft2_family_list_faces (PangoFontFamily  *family,
 			     PangoFontFace  ***faces,
@@ -841,32 +919,44 @@ pango_ft2_family_list_faces (PangoFontFamily  *family,
     {
       MiniXftFontSet *fontset;
       int i;
-      
-      fontset = MiniXftListFonts ((Display *)1, 0,
-				  XFT_ENCODING, MiniXftTypeString, "iso10646-1",
-				  XFT_FAMILY, MiniXftTypeString, ft2family->family_name,
-				  XFT_CORE, MiniXftTypeBool, False,
-				  NULL,
-				  XFT_STYLE,
-				  NULL);
-      
-      ft2family->n_faces = fontset->nfont;
-      ft2family->faces = g_new (PangoFT2Face *, ft2family->n_faces);
 
-      for (i = 0; i < fontset->nfont; i++)
+      if (is_alias_family (ft2family->family_name))
+  	{
+ 	  ft2family->n_faces = 4;
+ 	  ft2family->faces = g_new (PangoFT2Face *, ft2family->n_faces);
+ 
+ 	  i = 0;
+ 	  ft2family->faces[i++] = create_face (ft2family, "Regular");
+ 	  ft2family->faces[i++] = create_face (ft2family, "Bold");
+ 	  ft2family->faces[i++] = create_face (ft2family, "Italic");
+ 	  ft2family->faces[i++] = create_face (ft2family, "Bold Italic");
+	}
+      else
 	{
-	  char *s;
-	  MiniXftResult res;
+	  fontset = MiniXftListFonts ((Display *)1, 0,
+				      XFT_ENCODING, MiniXftTypeString, "iso10646-1",
+				      XFT_FAMILY, MiniXftTypeString, ft2family->family_name,
+				      XFT_CORE, MiniXftTypeBool, False,
+				      NULL,
+				      XFT_STYLE,
+				      NULL);
 	  
-	  res = MiniXftPatternGetString (fontset->fonts[i], XFT_STYLE, 0, &s);
-	  g_assert (res == MiniXftResultMatch);
+	  ft2family->n_faces = fontset->nfont;
+	  ft2family->faces = g_new (PangoFT2Face *, ft2family->n_faces);
 	  
-	  ft2family->faces[i] = g_object_new (PANGO_FT2_TYPE_FACE, NULL);
-	  ft2family->faces[i]->style = g_strdup (s);
-	  ft2family->faces[i]->family = ft2family;
+	  for (i = 0; i < fontset->nfont; i++)
+	    {
+	      char *s;
+	      MiniXftResult res;
+	      
+	      res = MiniXftPatternGetString (fontset->fonts[i], XFT_STYLE, 0, &s);
+	      g_assert (res == MiniXftResultMatch);
+	      
+	      ft2family->faces[i] = create_face (ft2family, s);
+	    }
+	  
+	  MiniXftFontSetDestroy (fontset);
 	}
-      
-      MiniXftFontSetDestroy (fontset);
     }
   
   if (n_faces)
